@@ -10,7 +10,7 @@ define('IMP_IMAPCLIENT_CONTINUATION', 2);
  *
  * TODO: This should eventually be moved to Horde 4.0/framework.
  *
- * $Horde: imp/lib/IMAP/Client.php,v 1.21.2.31 2007/01/02 13:54:58 jan Exp $
+ * $Horde: imp/lib/IMAP/Client.php,v 1.21.2.32 2007/07/10 09:59:00 slusarz Exp $
  *
  * Copyright 2005-2007 Michael Slusarz <slusarz@horde.org>
  *
@@ -586,8 +586,7 @@ class IMP_IMAPClient {
         );
 
         if ($this->queryCapability('NAMESPACE')) {
-            /*
-             * According to rfc2342 response from NAMESPACE command is:
+            /* According to RFC 2342, response from NAMESPACE command is:
              * * NAMESPACE (PERSONAL NAMESPACES) (OTHER_USERS NAMESPACE) (SHARED NAMESPACES)
              */
             $read = $this->_runCommand('NAMESPACE');
@@ -629,7 +628,9 @@ class IMP_IMAPClient {
                     $this->_namespace[$val] = array('name' => $val, 'delimiter' => $data[1], 'type' => $namespace_array[3], 'hidden' => true);
                 }
             }
-        } else {
+        }
+
+        if (empty($this->_namespace)) {
             $res = $this->_runCommand('LIST "" ""');
             if (is_a($res, 'PEAR_Error')) {
                 Horde::logMessage($res, __FILE__, __LINE__, PEAR_LOG_ERR);
