@@ -1,22 +1,22 @@
 <?php
 
 require_once 'Horde/MIME/Contents.php';
-require_once IMP_BASE . '/config/mime_drivers.php';
+// TODO: for BC - Remove for Horde 4.0
+require IMP_BASE . '/config/mime_drivers.php';
 
 /**
  * The IMP_Contents:: class extends the MIME_Contents:: class and contains
  * all functions related to handling the content and output of mail messages
  * in IMP.
  *
- * $Horde: imp/lib/MIME/Contents.php,v 1.153 2004/12/03 08:07:15 slusarz Exp $
+ * $Horde: imp/lib/MIME/Contents.php,v 1.153.4.45 2007/08/07 19:53:40 slusarz Exp $
  *
- * Copyright 2002-2004 Michael Slusarz <slusarz@bigworm.colorado.edu>
+ * Copyright 2002-2007 Michael Slusarz <slusarz@bigworm.colorado.edu>
  *
  * See the enclosed file COPYING for license information (GPL). If you
  * did not receive this file, see http://www.fsf.org/copyleft/gpl.html.
  *
  * @author  Michael Slusarz <slusarz@bigworm.colorado.edu>
- * @version $Revision: 1.153 $
  * @since   IMP 4.0
  * @package IMP
  */
@@ -25,35 +25,42 @@ class IMP_Contents extends MIME_Contents {
     /**
      * The text of the body of the message.
      *
-     * @var string $_body
+     * @var string
      */
     var $_body = '';
 
+    /**
+     * The MIME part id of the message body.
+     *
+     * @var integer
+     */
+    var $_body_id;
+
     /**
      * The text of various MIME body parts.
      *
-     * @var array $_bodypart
+     * @var array
      */
     var $_bodypart = array();
 
     /**
      * The IMAP index of the message.
      *
-     * @var integer $_index
+     * @var integer
      */
     var $_index;
 
     /**
-     * Are we currently in a print view?
+     * The mailbox of the current message.
      *
-     * @var boolean $_print
+     * @var string
      */
-    var $_print = false;
+    var $_mailbox;
 
     /**
      * Should attachment stripping links be generated?
      *
-     * @var boolean $_strip
+     * @var boolean
      */
     var $_strip = false;
 
@@ -67,11 +74,16 @@ class IMP_Contents extends MIME_Contents {
      * This method must be invoked as:
      *   $imp_contents = &IMP_Contents::singleton($index);
      *
-     * @access public
+     * @param string $index  The IMAP message mailbox/index. The index should
+     *                       be in the following format:
+     * <pre>
+     * msg_id IMP_IDX_SEP msg_folder
+     *   msg_id      = Message index of the message
+     *   IMP_IDX_SEP = IMP constant used to separate index/folder
+     *   msg_folder  = The full folder name containing the message index
+     * </pre>
      *
-     * @param integer $index  The IMAP message index.
-     *
-     * @return object IMP_Contents  The IMP_Contents object or null.
+     * @return IMP_Contents  The IMP_Contents object or null.
      */
     function &singleton($index)
     {
@@ -93,20 +105,21 @@ class IMP_Contents extends MIME_Contents {
     /**
      * Constructor
      *
-     * @access public
-     *
-     * @param mixed $in  Either the IMAP message index of the message to
-     *                   process, or a MIME_Message object.
+     * @param mixed $in  Either an index string (see IMP_Contents::singleton()
+     *                   for the format) or a MIME_Message object.
      */
     function IMP_Contents($in)
     {
         if (is_a($in, 'MIME_Message')) {
             $ob = $in;
         } else {
-            $this->_index = $in;
+            list($this->_index, $this->_mailbox) = explode(IMP_IDX_SEP, $in);
 
-            /* Get the MIME_Message object for the given index. */
-            $structure = @imap_fetchstructure($GLOBALS['imp']['stream'], $in, FT_UID);
+            /* Get the MIME_Structure object for the given index. */
+            require_once IMP_BASE . '/lib/IMAP.php';
+            $imp_imap = &IMP_IMAP::singleton();
+            $imp_imap->changeMbox($this->_mailbox);
+            $structure = @imap_fetchstructure($_SESSION['imp']['stream'], $this->_index, FT_UID);
             if (!is_object($structure)) {
                 return;
             }
@@ -134,14 +147,13 @@ class IMP_Contents extends MIME_Contents {
     /**
      * Returns the entire body of the message.
      *
-     * @access public
-     *
      * @return string  The text of the body of the message.
      */
     function getBody()
     {
         if (empty($this->_body)) {
-            $this->_body = @imap_body($GLOBALS['imp']['stream'], $this->_index, FT_UID);
+            $this->_body = @imap_body($_SESSION['imp']['stream'], $this->_index, FT_UID | FT_PEEK);
+            $this->_body = str_replace("\r\n", "\n", $this->_body);
         }
 
         return $this->_body;
@@ -150,28 +162,22 @@ class IMP_Contents extends MIME_Contents {
     /**
      * Gets the raw text for one section of the message.
      *
-     * @access public
-     *
      * @param integer $id  The ID of the MIME_Part.
      *
      * @return string  The text of the part.
      */
     function getBodyPart($id)
     {
-        $contents = '';
-
         if (!isset($this->_bodypart[$id])) {
-            $this->_bodypart[$id] = @imap_fetchbody($GLOBALS['imp']['stream'], $this->_index, $id, FT_UID);
+            $this->_bodypart[$id] = @imap_fetchbody($_SESSION['imp']['stream'], $this->_index, $id, FT_UID | FT_PEEK);
+            $this->_bodypart[$id] = str_replace("\r\n", "\n", $this->_bodypart[$id]);
         }
 
-        return ($contents . $this->_bodypart[$id]);
+        return isset($this->_bodypart[$id]) ? $this->_bodypart[$id] : '';
     }
 
     /**
-     * Allow attachments to be stripped by providing a link in summary
-     * view?
-     *
-     * @access public
+     * Allow attachments to be stripped by providing a link in summary view?
      *
      * @param boolean $strip  Should the strip links be generated?
      */
@@ -183,15 +189,14 @@ class IMP_Contents extends MIME_Contents {
     /**
      * Returns an array summarizing a part of a MIME message.
      *
-     * @access public
-     *
-     * @param object MIME_Part &$mime_part  See MIME_Contents::partSummary().
-     * @param optional boolean $guess       See MIME_Contents::partSummary().
+     * @param MIME_Part &$mime_part  See MIME_Contents::partSummary().
+     * @param boolean $guess         See MIME_Contents::partSummary().
      *
      * @return array  See MIME_Contents::partSummary().
      *                Adds the following key to that return:
      *                [6] = Compressed Download Link
-     *                [7] = Strip Attachment Link (if allowed)
+     *                [7] = Image Save link (if allowed)
+     *                [8] = Strip Attachment Link (if allowed)
      */
     function partSummary(&$mime_part, $guess = false)
     {
@@ -216,6 +221,18 @@ class IMP_Contents extends MIME_Contents {
             $summary[] = null;
         }
 
+        /* Display the image save link if the required registry calls are
+         * present. */
+        if (($mime_part->getPrimaryType() == 'image') &&
+            $registry->hasMethod('images/selectGalleries') &&
+            ($image_app = $registry->hasMethod('images/saveImage'))) {
+            Horde::addScriptFile('popup.js');
+            $url = Util::addParameter(Horde::applicationUrl('saveimage.php'), array('index' => ($this->_index . IMP_IDX_SEP . $this->_mailbox), 'id' => $mime_part->getMIMEId()));
+            $summary[] = Horde::link('#', _("Save Image in Gallery"), null, null, "popup_imp('" . $url . "',450,200); return false;") . '<img src="' . $registry->get('icon', $image_app) . '" alt="' . _("Save Image in Gallery") . '" title="' . _("Save Image in Gallery") . '" /></a>';
+        } else {
+            $summary[] = null;
+        }
+
         /* Strip the Attachment? */
         if ($this->_strip &&
             !$mime_part->getInformation('rfc822_part')) {
@@ -233,12 +250,10 @@ class IMP_Contents extends MIME_Contents {
     /**
      * Return the URL to the view.php page.
      *
-     * @access public
-     *
-     * @param object MIME_Part &$mime_part  See MIME_Contents::urlView().
-     * @param integer $actionID             See MIME_Contents::urlView().
-     * @param optional array $params        See MIME_Contents::urlView().
-     * @param optional boolean $dload       See MIME_Contents::urlView().
+     * @param MIME_Part &$mime_part  See MIME_Contents::urlView().
+     * @param integer $actionID      See MIME_Contents::urlView().
+     * @param array $params          See MIME_Contents::urlView().
+     * @param boolean $dload         See MIME_Contents::urlView().
      * The following parameter names will be overwritten by this function:
      *   id, index, mailbox
      *
@@ -249,10 +264,10 @@ class IMP_Contents extends MIME_Contents {
         require_once IMP_BASE . '/lib/Search.php';
 
         /* Add the necessary local parameters. */
-        $params = array_merge($params, IMP_Search::getSearchParameters(IMP::getThisMailbox()));
+        $params = array_merge($params, IMP::getSearchParameters($_SESSION['imp']['thismailbox']));
         $params['index'] = $this->_index;
         if (!isset($params['mailbox'])) {
-            $params['mailbox'] = $GLOBALS['imp']['mailbox'];
+            $params['mailbox'] = $_SESSION['imp']['mailbox'];
         }
 
         /* Should this be a download link? */
@@ -266,12 +281,10 @@ class IMP_Contents extends MIME_Contents {
     /**
      * Generate a link to the view.php page.
      *
-     * @access public
-     *
-     * @param object MIME_Part &$mime_part  See MIME_Contents::linkView().
-     * @param integer $actionID             See MIME_Contents::linkView().
-     * @param string $text                  See MIME_Contents::linkView().
-     * @param optional array $params        See MIME_Contents::linkView().
+     * @param MIME_Part &$mime_part  See MIME_Contents::linkView().
+     * @param integer $actionID      See MIME_Contents::linkView().
+     * @param string $text           See MIME_Contents::linkView().
+     * @param array $params          See MIME_Contents::linkView().
      *
      * @return string  See MIME_Contents::linkView().
      */
@@ -288,7 +301,7 @@ class IMP_Contents extends MIME_Contents {
 
         /* If download attachment, add the 'thismailbox' param. */
         if ($actionID == 'download_attach') {
-            $params['viewparams']['thismailbox'] = IMP::getThisMailbox();
+            $params['viewparams']['thismailbox'] = $_SESSION['imp']['thismailbox'];
         }
 
         if ($mime_part->getInformation('viewparams')) {
@@ -301,60 +314,31 @@ class IMP_Contents extends MIME_Contents {
     }
 
     /**
-     * Returns the full message text for a given message/mailbox.
-     *
-     * @access public
-     *
-     * @param optional integer $index  The index/mailbox of the message. If
-     *                                 empty, will use the current message
-     *                                 index.
+     * Returns the full message text.
      *
      * @return string  The full message text.
      */
-    function fullMessageText($index = null)
+    function fullMessageText()
     {
-        $imp_headers = &$this->getHeaderOb($index);
+        $imp_headers = &$this->getHeaderOb();
         return $imp_headers->getHeaderText() . $this->getBody();
     }
 
     /**
-     * Returns the header object for the given message.
+     * Returns the header object.
      *
-     * @access public
-     * @param optional integer $index  The index/mailbox of the message. If
-     *                                 empty, will use the current message
-     *                                 index.
-     *
-     * @return object IMP_Headers  The IMP_Headers object.
+     * @return IMP_Headers  The IMP_Headers object.
      */
-    function &getHeaderOb($index = null)
+    function &getHeaderOb()
     {
-        if (is_null($index)) {
-            $index = $this->_index;
-        } elseif ($index != $this->_index) {
-            if (strstr($index, IMP_IDX_SEP)) {
-                require_once IMP_BASE . '/lib/base.php';
-                $imp_imap = &IMP_IMAP::singleton();
-                list($index, $mailbox) = explode(IMP_IDX_SEP, $index);
-                $imp_imap->changeMbox($mailbox);
-            }
-            if ($index != $this->_index) {
-                $this->_body = '';
-                $this->_index = $index;
-            }
-        }
-
         require_once IMP_BASE . '/lib/MIME/Headers.php';
-        $imp_headers = &new IMP_Headers($index);
-
+        $imp_headers = &new IMP_Headers($this->_index);
         return $imp_headers;
     }
 
     /**
      * Returns the IMAP index for the current message.
      *
-     * @access public
-     *
      * @return integer  The message index.
      */
     function getMessageIndex()
@@ -364,19 +348,20 @@ class IMP_Contents extends MIME_Contents {
 
     /**
      * Rebuild the MIME_Part structure of a message from IMAP data.
+     * This will store IMAP data in all parts of the message - for example,
+     * all data for a multipart/mixed part will be stored in the base part,
+     * and each part will contain its own data.  Note that if you want to
+     * build a message string from the MIME_Part data after calling
+     * rebuildMessage(), you should use IMP_Contents::toString() instead of
+     * MIME_Part::toString().
      *
-     * @access public
-     *
-     * @param optional boolean $decode  Decode the body parts?
-     *
-     * @return object MIME_Message  A MIME_Message object with all of the body
-     *                              text stored in the individual MIME_Parts.
+     * @return MIME_Message  A MIME_Message object with all of the body text
+     *                       stored in the individual MIME_Parts.
      */
-    function rebuildMessage($decode = false)
+    function rebuildMessage()
     {
         $part = $this->_message->getBasePart();
-        $this->_rebuildMessage($part, $decode);
-
+        $this->_rebuildMessage($part);
         return $this->_message;
     }
 
@@ -386,28 +371,24 @@ class IMP_Contents extends MIME_Contents {
      *
      * @access private
      *
-     * @param object MIME_Part &$part  A MIME_Part object.
-     * @param boolean $decode          Decode the body parts?
+     * @param MIME_Part $part  A MIME_Part object.
      */
-    function _rebuildMessage(&$part, $decode)
+    function _rebuildMessage($part)
     {
         $id = $part->getMIMEId();
 
+        $mime_part = $this->getMIMEPart($id);
+        $this->_setContents($mime_part, true);
+        if ($this->_message->getPart($id) === $this->_message) {
+            $this->_message = &$mime_part;
+        } else {
+            $this->_message->alterPart($id, $mime_part);
+        }
+
         if ($part->getPrimaryType() == 'multipart') {
             /* Recursively process any subparts. */
             foreach ($part->getParts() as $mime) {
-                $this->_rebuildMessage($mime, $decode);
-            }
-        } elseif ($part->getType() == 'message/rfc822') {
-            $part->setContents($this->getBodyPart(substr($id, 0, strrpos($id, '.'))));
-            $this->_message->alterPart($id, $part);
-        } else {
-            if ($decode) {
-                $this->_message->alterPart($id, $this->getDecodedMIMEPart($id));
-            } else {
-                $mime_part = &$this->getMIMEPart($id);
-                $mime_part->setContents($this->getBodyPart($id));
-                $this->_message->alterPart($id, $mime_part);
+                $this->_rebuildMessage($mime);
             }
         }
     }
@@ -415,16 +396,13 @@ class IMP_Contents extends MIME_Contents {
     /**
      * Render a MIME Part.
      *
-     * @param object MIME_Part &$mime_part  A MIME_Part object.
+     * @param MIME_Part &$mime_part  A MIME_Part object.
      *
      * @return string  The rendered data.
      */
     function renderMIMEPart(&$mime_part)
     {
-        if (!$mime_part->getContents()) {
-            $mime_part->setContents($this->getBodyPart($mime_part->getMIMEId()));
-        }
-
+        $this->_setContents($mime_part);
         $text = parent::renderMIMEPart($mime_part);
 
         /* Convert textual emoticons into graphical ones - but only for
@@ -432,7 +410,7 @@ class IMP_Contents extends MIME_Contents {
         if (($mime_part->getPrimaryType() == 'text') &&
             $GLOBALS['prefs']->getValue('emoticons')) {
             require_once 'Horde/Text/Filter.php';
-            $text = Text_Filter::filter($text, 'emoticons', array(true));
+            $text = Text_Filter::filter($text, 'emoticons', array('entities' => true));
         }
 
         return $text;
@@ -445,12 +423,31 @@ class IMP_Contents extends MIME_Contents {
      */
     function _addCacheShutdown()
     {
-        /* Don't save the cached $_body or $_bodypart arrays since we most
-           likely will not use them in the cached object (since the cache
-           will generally be used to view parts that _weren't* displayed
-           originally. */
+        /* Don't cache the $_bodypart array since we most likely will not use
+         * them in the cached object (since the cache will generally be used
+         * to view parts that weren't displayed originally).
+         * However, we will lose the benefit of not having to
+         * make calls to the mail server on subsequent page loads, so allow
+         * users to change this behavior.  Only cache small messages as
+         * storing large messages will make the session size too large and
+         * will tax storage and performance on the server. */
+        if (empty($GLOBALS['conf']['server']['cache_msgbody'])) {
+            $this->_bodypart = array();
+        } else {
+            $size = 0;
+            foreach ($this->_bodypart as $key => $val) {
+                $part_size = strlen($val);
+                if (($size + $part_size) < 10240) {
+                    $size += $part_size;
+                } else {
+                    unset($this->_bodypart[$key]);
+                }
+            }
+        }
+
+        /* Never cache the full body part as we use this item much less
+         * often than specific message body parts. */
         $this->_body = null;
-        $this->_bodypart = array();
 
         parent::_addCacheShutdown();
     }
@@ -458,8 +455,6 @@ class IMP_Contents extends MIME_Contents {
     /**
      * Get the from address of the message.
      *
-     * @access public
-     *
      * @return string  The from address of the message.
      */
     function getFromAddress()
@@ -472,29 +467,53 @@ class IMP_Contents extends MIME_Contents {
     /**
      * Generate the list of MIME IDs to use for download all.
      *
-     * @access public
+     * @param boolean $forward  Generate a list of items that can be forwarded,
+     *                          rather than downloaded.
      *
      * @return array  The list of MIME IDs that should be downloaded when
      *                downloading all attachments.
      */
-    function getDownloadAllList()
+    function getDownloadAllList($forward = false)
     {
         $downloads = array();
 
-        /* Here is what we don't consider 'downloadable':
-           Any 'multipart/*' part
-           All 'message/*' parts except for 'message/rfc822'
-           The first viewable part */
+        $bodyid = $this->findBody();
+
+        /* Here is what we consider 'forwardable':
+         * All parts > 0
+         * Part 0 if not multipart/mixed
+         * NOT the body part (if one exists)
+         *
+         * Here is what we consider 'downloadable':
+         * All parts not 'multipart/*' and 'message/*' except for
+         *  'message/rfc822'
+         * All parts that are not PGP or S/MIME signature information
+         * NOT the body part (if one exists) */
         foreach ($this->_message->contentTypeMap() as $key => $val) {
-            if (($key == 0) && (strstr($val, 'multipart/mixed') === false)) {
-                break;
+            if ($key == $bodyid) {
+                continue;
             }
 
-            if ((intval($key) != 1) &&
-                (strstr($val, '1.') === false) &&
-                (strstr($val, 'multipart/') === false) &&
-                (strstr($val, 'message/') === false)) {
-                $downloads[] = $key;
+            if ($forward) {
+                if ($key == 0) {
+                    if ($val != 'multipart/mixed') {
+                        $downloads[] = $key;
+                        break;
+                    }
+                } elseif (strpos($key, '.') === false) {
+                    $downloads[] = $key;
+                }
+            } else {
+                if (strpos($val, 'message/') === 0) {
+                    if (strpos($val, 'message/rfc822') === 0) {
+                        $downloads[] = $key;
+                    }
+                } elseif (((intval($key) != 1) && (strpos($val, '1.') === false)) &&
+                          ((strpos($val, 'multipart/') === false)) &&
+                          (strpos($val, 'application/x-pkcs7-signature') === false) &&
+                          (strpos($val, 'application/pkcs7-signature') === false)) {
+                    $downloads[] = $key;
+                }
             }
         }
 
@@ -504,8 +523,6 @@ class IMP_Contents extends MIME_Contents {
     /**
      * Generate a download all link, if possible.
      *
-     * @access public
-     *
      * @return string  The download link.
      */
     function getDownloadAllLink()
@@ -524,45 +541,18 @@ class IMP_Contents extends MIME_Contents {
         return $url;
     }
 
-    /**
-     * Sets the print mode.
-     *
-     * @access public
-     *
-     * @param boolean $mode  True = print mode; False = non-print mode.
-     */
-    function setPrintMode($mode)
-    {
-        $this->_print = $mode;
-    }
-
-    /**
-     * Gets the print mode.
-     *
-     * @access public
-     *
-     * @return boolean  True = print mode; False = non-print mode.
-     */
-    function getPrintMode()
-    {
-        return $this->_print;
-    }
-
     /**
      * Prints out the status message for a given MIME Part.
      *
-     * @access public
-     *
-     * @param mixed $msg               See MIME_Contents::formatStatusMsg().
-     * @param optional string $img     See MIME_Contents::formatStatusMsg().
-     * @param optional boolean $print  Output this message when in a print
-     *                                 view?
+     * @param mixed $msg      See MIME_Contents::formatStatusMsg().
+     * @param string $img     See MIME_Contents::formatStatusMsg().
+     * @param boolean $print  Output this message when in a print view?
      *
      * @return string  The formatted status message.
      */
     function formatStatusMsg($msg, $img = null, $printable = true)
     {
-        if (!$printable && $this->_print) {
+        if (!$printable && IMP::printMode()) {
             return '';
         } else {
             return parent::formatStatusMsg($msg, $img);
@@ -571,26 +561,30 @@ class IMP_Contents extends MIME_Contents {
 
     /**
      * Finds the main "body" text part (if any) in a message.
-     *
-     * @access public
+     * "Body" data is the first text part in the base MIME part.
      *
      * @return string  The MIME ID of the main "body" part.
      */
     function findBody()
     {
-        /* Look for potential body parts. */
-        $mime_message = $this->rebuildMessage();
-        $part = $mime_message->getBasePart();
-        $primary_type = $part->getPrimaryType();
-        if (($primary_type == MIME::type(TYPEMULTIPART)) ||
-            ($primary_type == MIME::type(TYPETEXT))) {
-            return $this->_findBody($part);
+        if (isset($this->_body_id)) {
+            return $this->_body_id;
+        }
+
+        if (!is_null($this->_message)) {
+            /* Look for potential body parts. */
+            $part = $this->_message->getBasePart();
+            $primary_type = $part->getPrimaryType();
+            if (($primary_type == MIME::type(TYPEMULTIPART)) ||
+                ($primary_type == MIME::type(TYPETEXT))) {
+                $this->_body_id = $this->_findBody($part);
+                return $this->_body_id;
+            }
         }
 
         return null;
     }
 
-
     /**
      * Processes a MIME Part and looks for "body" data.
      *
@@ -598,18 +592,21 @@ class IMP_Contents extends MIME_Contents {
      *
      * @return string  The MIME ID of the main "body" part.
      */
-    function _findBody(&$mime_part)
+    function _findBody($mime_part)
     {
-        if ($mime_part->getPrimaryType() == MIME::type(TYPEMULTIPART)) {
-            foreach ($mime_part->getParts() as $part) {
-                if (($partid = $this->_findBody($part))) {
-                    return $partid;
+        if (intval($mime_part->getMIMEId()) < 2 ||
+            $mime_part->getInformation('alternative') === 0) {
+            if ($mime_part->getPrimaryType() == MIME::type(TYPEMULTIPART)) {
+                foreach ($mime_part->getParts() as $part) {
+                    if (($partid = $this->_findBody($part))) {
+                        return $partid;
+                    }
+                }
+            } elseif ($mime_part->getPrimaryType() == MIME::type(TYPETEXT)) {
+                if ($mime_part->getBytes() ||
+                    $this->getBodyPart($mime_part->getMIMEId())) {
+                    return $mime_part->getMIMEId();
                 }
-            }
-        } elseif ($mime_part->getPrimaryType() == MIME::type(TYPETEXT)) {
-            if ($mime_part->getContents() ||
-                $this->getBodyPart($mime_part->getMIMEId())) {
-                return $mime_part->getMIMEId();
             }
         }
 
@@ -621,7 +618,7 @@ class IMP_Contents extends MIME_Contents {
      *
      * @access private
      *
-     * @param optional integer $index  The IMAP index of the current message.
+     * @param integer $index  The IMAP index of the current message.
      *
      * @return string  A unique cache ID.
      */
@@ -634,7 +631,83 @@ class IMP_Contents extends MIME_Contents {
                 return parent::_createCacheID();
             }
         }
-        return md5(implode('|', array($index, $GLOBALS['imp']['mailbox'])));
+        return md5(implode('|', array($index, $_SESSION['imp']['thismailbox'])));
+    }
+
+    /**
+     * Make sure the contents of the current part are set from IMAP server
+     * data.
+     *
+     * @access private
+     * @since IMP 3.1
+     *
+     * @param MIME_Part &$mime_part  The MIME_Part object to work with.
+     * @param boolean $all           Download the entire parts contents?
+     */
+    function _setContents(&$mime_part, $all = false)
+    {
+        if (!$mime_part->getInformation('imp_contents_set') &&
+            !$mime_part->getContents()) {
+            $id = $mime_part->getMIMEId();
+            if ($all && ($mime_part->getType() == 'message/rfc822')) {
+                $id = substr($mime_part->getMIMEId(), 0, -2);
+            }
+            $mime_part->setContents($this->getBodyPart($id));
+        }
+        $mime_part->setInformation('imp_contents_set', true);
+    }
+
+    /**
+     * Fetch part of a MIME message.
+     *
+     * @since IMP 3.1
+     *
+     * @param integer $id   The MIME ID of the part requested.
+     * @param boolean $all  If this is a header part, should we return all text
+     *                      in the body?
+     *
+     * @return MIME_Part  The MIME_Part.
+     */
+    function &getRawMIMEPart($id, $all = false)
+    {
+        $mime_part = $this->getMIMEPart($id);
+        if (!is_a($mime_part, 'MIME_Part')) {
+            $mime_part = null;
+            return $mime_part;
+        }
+        $this->_setContents($mime_part, true);
+        return $mime_part;
+    }
+
+    /**
+     * Create a message string from a MIME message that has used
+     * rebuildMessage() to build the data from the IMAP server.
+     *
+     * @since IMP 4.1
+     *
+     * @param MIME_Message $message  A MIME_Message object.
+     * @param boolean $canonical     Return a canonical string?
+     *
+     * @return string  The contents of the MIME_Message object.
+     */
+    function toString($message, $canonical = false)
+    {
+        $text = '';
+
+        $part = $message->getBasePart();
+        foreach ($part->contentTypeMap() as $key => $val) {
+            if (($key != 0) && (strpos($val, 'multipart/') === 0)) {
+                $old = $part->getPart($key);
+                $old->setContents('');
+                $part->alterPart($key, $old);
+            }
+        }
+
+        if ($message->getMIMEId() == 0) {
+            $part->setContents('');
+        }
+
+        return ($canonical) ? $part->toCanonicalString() : $part->toString();
     }
 
 }
